/**********************************************************************
 *<
	FILE: SFCMODimp.cpp

	DESCRIPTION:  Starfleet Command file import module

	CREATED BY: Sean Dumas

	HISTORY: created 16 December 1999

 *>	Copyright (c) 1999, All Rights Reserved.
 **********************************************************************/

#include "Max.h"
#include <fstream.h>
#include <stdio.h>
#include <direct.h>
#include <commdlg.h>
#include <io.h>
#include <vector>
#include "imtl.h"
#include "dummy.h"
#include "istdplug.h"
#include "stdmat.h"
#include "gamma.h"
#include "helpsys.h"

#include "SFCMODImpResource.h"
#include "SFCMODImpExp.h"

using namespace std;

const double kPI = 3.14159265;
const int vIdx[3] = { 0, 2, 1 };

void debugMsg( char *string, ... )
{
	FILE *debugFile;
	static bool firstTime = TRUE;
	char outString[ 100 ];
	va_list argList;

#ifndef _DEBUG
	return;
#endif

	if ( firstTime )
	{
		debugFile = fopen( "debug.log", "wb" );
		firstTime = FALSE;
	}
	else
	{
		debugFile = fopen( "debug.log", "ab" );
	}

		// Get the variable argument list
	va_start( argList, string );
	vsprintf( outString, string, argList );
	va_end( argList );

	strcat( outString, "\r\n" );
	fprintf( debugFile, outString );

	fclose( debugFile );
}


#pragma pack()


struct tPoint3
{
	float x, y, z;
};
class tSFCMODImport : public SceneImport 
{
private:
	void mReadChunk( FILE* myFile, const fpos_t kEndPos );
	void mSplitBreakUpMeshes();
	INode* mFindNode( INode* pCurrNode, Class_ID& id );
	void mBuildScene( ImpInterface* ii, Interface* gi );
	int mLoadSFCMOD( const TCHAR* filename, ImpInterface* ii, Interface* gi );

private:
	enum
	{
		kMaxLOD = 5
	};
#pragma pack(1)
	struct tFlatHeader
	{
		unsigned int fVersion;
		float fRadius;
		int fTotalLODs;
		float fLODHysteresis;
	}fFlatHeader;

	struct tNamedPoint
	{
		tPoint3			fPos;
		unsigned int	fNameOffset;
	};

	struct tFaceDef
	{
		int vtxCount;
		int materialNdx;
		tPoint3 normal;
		struct tFlatPolyVertex
		{
			unsigned short vtxNdx;
			unsigned short flags;
			tPoint3 normal;
			float u;
			float v;
			float light;
		}*vertices;
		tFaceDef() : vertices(NULL) {}
		~tFaceDef() { if ( vertices ) { delete [] vertices; } }
	};

	struct tMaterial
	{
		enum
		{
			kTexMapped      = 0x0001,	// set if texture mapping
			kGouraud        = 0x0002,	// set if gouraud (smooth) shading
			kFlat			= 0x0004,	// set if flat shading
			kWire           = 0x0008,	// set if polygon should be wire frame
			kAdditive		= 0x0010,	// set if polygon should be blended
			kLuminance		= 0x0020,	// set if we have luminance texture maps
		};
		unsigned int fFlags;
		COLORREF fColor;
		float fAmbient;
		float fDiffuse;
		float fEmittance;
		struct tTextureInfo
		{
			int fTexID;
			int fTexNameOffset;
		};
		tTextureInfo fDiffuseTexture;
		tTextureInfo fLuminanceTexture;
	};
#pragma pack()

	float fLODTransitions[kMaxLOD];

	tNamedPoint* fNamedPoints;
	int fNumNamedPoints;

	Tab<int> fNameIndices;
	Tab<char> fStringPool;

	tMaterial* fMaterials;
	int fNumMaterials;

	class tMesh
	{
	public:
		tPoint3*		fVertices;
		int				fNumVertices;
		tFaceDef*		fFaces;
		unsigned int*	fSmoothing;
		int				fNumFaces;
		
		tMesh* fNext;

		tMesh();
		~tMesh();
		void mCalcNormal(tFaceDef& face);
		void mCalcVertexNormals();
	}*fFirstMesh;

	char fFilename[_MAX_FNAME];
	char fDrive[_MAX_DRIVE];
	char fDirectory[_MAX_DIR];

public:
					tSFCMODImport();
					~tSFCMODImport();
	int				ExtCount();					// Number of extensions supported
	const TCHAR *	Ext(int n);					// Extension #n (i.e. "MOD")
	const TCHAR *	LongDesc();					// Long ASCII description (i.e. "Starfleet Command model file")
	const TCHAR *	ShortDesc();				// Short ASCII description (i.e. "Starfleet Command")
	const TCHAR *	AuthorName();				// ASCII Author name
	const TCHAR *	CopyrightMessage();			// ASCII Copyright message
	const TCHAR *	OtherMessage1();			// Other message #1
	const TCHAR *	OtherMessage2();			// Other message #2
	unsigned int	Version();					// Version number * 100 (i.e. v3.01 = 301)
	void			ShowAbout(HWND hWnd);		// Show DLL's "About..." box
	int				DoImport(const TCHAR *name,ImpInterface *i,Interface *gi, BOOL suppressPrompts=FALSE);	// Import file
private:
	void			mReadMaterial( tMaterial* material, FILE* file );
	fpos_t			mGetFilePos( FILE* file );
};

// Handy file class


// Node list structure


struct MtlName
{
	char s[20];
};

//
// Beginning of old file
//

//#define PRINTCHUNKID

HINSTANCE hInstance;

static BOOL showPrompts;

TCHAR *GetString(int id)
{
	static TCHAR buf[256];
	if (hInstance)
		return LoadString(hInstance, id, buf, sizeof(buf)) ? buf : NULL;
	return NULL;
}


static int MessageBox(int s1, int s2, int option = MB_OK) 
{
	TSTR str1(GetString(s1));
	TSTR str2(GetString(s2));
	return MessageBox(GetActiveWindow(), str1, str2, option);
}

static int MessageBox2(int s1, int s2, int option = MB_OK) 
{
	TSTR str1(GetString(s1));
	TSTR str2(GetString(s2));
	return MessageBox(GetActiveWindow(), str1, str2, option);
}

static int MessageBox(TCHAR *s, int s2, int option = MB_OK) 
{
	TSTR str1(s);
	TSTR str2(GetString(s2));
	return MessageBox(GetActiveWindow(), str1, str2, option);
}


//===========================================================


// Handy pointers to importers
tSFCMODImport *theImport = NULL;

// Dialogue message proceedure
BOOL CALLBACK MsgDlgProc(HWND hDlg, UINT iMsg, UINT wParam, LONG lParam)
{
    switch (iMsg)
    {
	    case WM_INITDIALOG:
			CenterWindow(hDlg, GetWindow(hDlg, GW_OWNER));
			return TRUE;

	    case WM_COMMAND:
			switch(LOWORD(wParam))
            {
				case IDOK:	
					EndDialog(hDlg, TRUE); 
					break;
			}
			break;
	}
    return FALSE;
}

// Jaguar interface code

int controlsInit = FALSE;

BOOL WINAPI DllMain(HINSTANCE hinstDLL,ULONG fdwReason,LPVOID lpvReserved) 
{
	hInstance = hinstDLL;

	if ( !controlsInit ) 
	{
		controlsInit = TRUE;
		
		// jaguar controls
		InitCustomControls(hInstance);

		// initialize Chicago controls
		InitCommonControls();
	}
	switch(fdwReason) 
	{
		case DLL_PROCESS_ATTACH:
			//MessageBox(NULL,_T("3DSIMP.DLL: DllMain"),_T("3DSIMP"),MB_OK);
			break;
		case DLL_THREAD_ATTACH:
			break;
		case DLL_THREAD_DETACH:
			break;
		case DLL_PROCESS_DETACH:
			break;
	}
	return(TRUE);
}


//------------------------------------------------------

class tSFCMODClassDesc:public ClassDesc 
{
	public:
	int 			IsPublic() { return 1; }
	void *			Create(BOOL loading = FALSE) { return new tSFCMODImport; }
	const TCHAR *	ClassName() { return GetString(IDS_SD_SFC); }
	SClass_ID		SuperClassID() { return SCENE_IMPORT_CLASS_ID; }
	Class_ID		ClassID() { return Class_ID(0x1e534ad7, 0x2417a8f); }
	const TCHAR* 	Category() { return GetString(IDS_SD_IMPORT);  }
};

static tSFCMODClassDesc tSFCMODDesc;

// Statics

//------------------------------------------------------
// This is the interface to Jaguar:
//------------------------------------------------------

__declspec( dllexport ) const TCHAR *LibDescription() 
{ 
	return GetString(IDS_SD_SFCMODIMPORTDLL); 
}

__declspec( dllexport ) int LibNumberClasses() 
{ 
	return 1; 
}

__declspec( dllexport ) ClassDesc *LibClassDesc(int i) 
{
	switch(i) 
	{
		case 0: return &tSFCMODDesc; break;
		default: return 0; break;
	}
}

// Return version so can detect obsolete DLLs
__declspec( dllexport ) ULONG LibVersion() 
{ 
	return VERSION_3DSMAX; 
}

// Let the plug-in register itself for deferred loading
__declspec( dllexport ) ULONG CanAutoDefer()
{
	return 1;
}

//
// SFCMOD import module functions follow:
//

tSFCMODImport::tMesh::tMesh()	:
	fVertices(NULL),
	fFaces(NULL),
	fSmoothing(NULL),
	fNumVertices(0),
	fNumFaces(0),
	fNext(NULL)
{
}

tSFCMODImport::tMesh::~tMesh()
{
	if ( fVertices )
	{
		delete [] fVertices;
	}

	if ( fFaces )
	{
		delete [] fFaces;
	}

	if ( fSmoothing )
	{
		delete [] fSmoothing;
	}
}

tSFCMODImport::tSFCMODImport() :
	fNamedPoints(NULL),
	fNumNamedPoints(0),
	fFirstMesh(NULL),
	fMaterials(NULL),
	fNumMaterials(0)
{
}

tSFCMODImport::~tSFCMODImport() 
{
	if ( fNamedPoints )
	{
		delete [] fNamedPoints;
	}

	tMesh* mesh = fFirstMesh;
	while ( mesh )
	{
		tMesh* oldMesh = mesh;
		mesh = mesh->fNext;
		delete oldMesh;
	}

	if ( fMaterials )
	{
		delete [] fMaterials;
	}
}

int tSFCMODImport::ExtCount() 
{
	return 1;
}

const TCHAR *tSFCMODImport::Ext(int n) 
{		// Extensions supported for import/export modules
	switch(n) 
	{
		case 0:
			return _T("MOD");
	}
	return _T("");
}

const TCHAR *tSFCMODImport::LongDesc() 
{			// Long ASCII description (i.e. "Targa 2.0 Image File")
	return GetString(IDS_SD_SFCMODFILE);
}
	
const TCHAR *tSFCMODImport::ShortDesc() 
{			// Short ASCII description (i.e. "Targa")
	return GetString(IDS_SD_SFCFILE);
}

const TCHAR *tSFCMODImport::AuthorName() 
{			// ASCII Author name
	return GetString(IDS_SD_AUTHOR);	// Hey, that's me!!!!
}

const TCHAR *tSFCMODImport::CopyrightMessage() 
{	// ASCII Copyright message
	return GetString(IDS_SD_COPYRIGHT);
}

const TCHAR *tSFCMODImport::OtherMessage1() 
{		// Other message #1
	return _T("");
}

const TCHAR *tSFCMODImport::OtherMessage2() 
{		// Other message #2
	return _T("");
}

unsigned int tSFCMODImport::Version() 
{				// Version number * 100 (i.e. v3.01 = 301)
	return 200;
}

void tSFCMODImport::ShowAbout(HWND hWnd) 
{			// Optional

}

void tSFCMODImport::mReadMaterial( tMaterial* material, FILE* file )
{
	if ( !material )
	{
		return;
	}

	fread( &material->fFlags, sizeof(unsigned int), 1, file );
	fread( &material->fColor, sizeof(COLORREF), 1, file );
	fread( &material->fAmbient, sizeof(float), 1, file );
	fread( &material->fDiffuse, sizeof(float), 1, file );
	fread( &material->fEmittance, sizeof(float), 1, file );
	fread( &material->fDiffuseTexture, sizeof(tMaterial::tTextureInfo), 1, file );

	if ( material->fFlags & tMaterial::kLuminance )
	{
		fread( &material->fLuminanceTexture, sizeof(tMaterial::tTextureInfo), 1, file );
	}
}

fpos_t tSFCMODImport::mGetFilePos( FILE* file )
{
	fpos_t pos;
	fgetpos( file, &pos );
	return pos;
}

void tSFCMODImport::mReadChunk( FILE* myFile, const fpos_t kEndPos )
{
	static tMesh* pCurMesh = NULL;
	static tFaceDef* pCurFace = NULL;

	while ( !feof(myFile) )
	{
		fpos_t pos;
		fgetpos ( myFile, &pos );
		if ( pos >= kEndPos )
		{
			return;
		}
		struct tChunk
		{
			FOURCC	fType;
			int		fSize;
		}chunk;
		fread( &chunk, sizeof(tChunk), 1, myFile );
		fgetpos( myFile, &pos );

		switch( chunk.fType )
		{
		case MAKEFOURCC('M', 'D', 'L', 'S'):
			{
				debugMsg( "Model file main Chunk" );
				fread( &fFlatHeader, sizeof( tFlatHeader ), 1, myFile );
				fread( &fLODTransitions, sizeof( float ), kMaxLOD, myFile );
				mReadChunk( myFile, pos + chunk.fSize );
			}
			break;
		case MAKEFOURCC('S','T','R','S'):
			debugMsg( "Strings chunk" );
			fStringPool.Resize(chunk.fSize);
			fStringPool.SetCount(chunk.fSize);
			fread( fStringPool.Addr(0), sizeof(char), chunk.fSize, myFile );
			break;
		case MAKEFOURCC('M','A','T','S'):
			{
				debugMsg( "Materials chunk" );
				// Materials are no longer uniform in size.  Some may contain luminance maps others may not
				vector<tMaterial> materials;
				const fpos_t kStart = mGetFilePos(myFile);	// be back

				while ( mGetFilePos(myFile) - kStart < chunk.fSize )
				{
					tMaterial material;
					mReadMaterial( &material, myFile );
					materials.push_back(material);
				}
				// No copy our vector into the array
				fNumMaterials = materials.size();
				if ( fNumMaterials )
				{
					fMaterials	= new tMaterial[fNumMaterials];
				}
				else
				{
					fMaterials = NULL;
				}
				if ( fMaterials )
				{
					for ( int i = 0; i < materials.size(); i++ )
					{
						fMaterials[i] = materials[i];
					}
				}
			}
			break;
		case MAKEFOURCC('P','N','T','S'):
			debugMsg( "Named points chunk" );
			fNumNamedPoints = chunk.fSize / sizeof(tNamedPoint);
			if ( fNumNamedPoints )
			{
				fNamedPoints = new tNamedPoint[fNumNamedPoints];
				fread( fNamedPoints, sizeof(tNamedPoint), fNumNamedPoints, myFile );
			}
			break;

		//
		//
		case MAKEFOURCC('M','O','D','0'):
			{
				debugMsg( "Geometry chunk" );
				// Create a new mesh
				pCurMesh = new tMesh;
				if ( fFirstMesh )
				{
					tMesh* pMesh = fFirstMesh;
					while ( pMesh->fNext )
					{
						pMesh = pMesh->fNext;
					}
					pMesh->fNext = pCurMesh;
				}
				else
				{
					fFirstMesh = pCurMesh;
				}
				pCurMesh->fNext = NULL;

				fread( &pCurMesh->fNumVertices, sizeof(int), 1, myFile );
				fread( &pCurMesh->fNumFaces, sizeof(int), 1, myFile );
				pCurMesh->fVertices = new tPoint3[pCurMesh->fNumVertices];
				pCurMesh->fFaces = new tFaceDef[pCurMesh->fNumFaces];
				pCurFace = pCurMesh->fFaces;

				mReadChunk( myFile, pos + chunk.fSize );
			}
			break;
		case MAKEFOURCC('V', 'T', 'X', 'S' ):
			debugMsg( "Vertex chunk" );
			if ( pCurMesh )
			{
				fread( pCurMesh->fVertices, sizeof(tPoint3), pCurMesh->fNumVertices, myFile );
			}
			break;
		case MAKEFOURCC('P', 'O', 'L', 'Y' ):
			// debugMsg( "Polygon Chunk" );
			if ( pCurMesh && pCurFace )
			{
				fread( &pCurFace->vtxCount, sizeof(int), 1, myFile );
				fread( &pCurFace->materialNdx, sizeof(int), 1, myFile );
				fread( &pCurFace->normal, sizeof(tPoint3), 1, myFile );
				pCurFace->vertices = new tFaceDef::tFlatPolyVertex[pCurFace->vtxCount];
				fread( pCurFace->vertices, sizeof(tFaceDef::tFlatPolyVertex), pCurFace->vtxCount, myFile );
				pCurFace++;
			}
			break;

		default:
			debugMsg( "Unknown or invalid chunk" );
			break;
		}

		// Advance to end of data (which is start of next chunk)
		// Note: Data blocks are padded to end on a dword boundary
		//
		const fpos_t kNextPos = ((pos + chunk.fSize) + 3) & ~3;
		fsetpos( myFile, &kNextPos );
	}
}

INode* tSFCMODImport::mFindNode( INode* pCurrNode, Class_ID& id )
{
	Object *pObject = pCurrNode->GetObjectRef();
	if ( pObject && pObject->ClassID() == id )
	{
		return pCurrNode;
	}
	else
	{
		for ( int i = 0; i < pCurrNode->NumberOfChildren(); i++ )
		{
			INode* pNode = mFindNode( pCurrNode->GetChildNode(i), id );
			if ( pNode )
			{
				return pNode;
			}
		}
		return NULL;
	}
}

void tSFCMODImport::mBuildScene( ImpInterface* ii, Interface* gi )
{
	int i;

	// Create the main Multi-Material
	char string[ _MAX_PATH + 1 ];
	MultiMtl* pMultiMtl = NewDefaultMultiMtl();
	sprintf( string, "%s_Material", fFilename );
	pMultiMtl->SetName( string );

	// Create all the sub-materials
	if ( fNumMaterials )
	{
		pMultiMtl->SetNumSubMtls(fNumMaterials);
		for ( i = 0; i < fNumMaterials; i++ )
		{
			StdMat2* pMtl = NewDefaultStdMat();

			pMtl->SetAmbient( Color( fMaterials[i].fAmbient,  fMaterials[i].fAmbient, fMaterials[i].fAmbient ), 0 );
			pMtl->SetDiffuse( Color( fMaterials[i].fDiffuse, fMaterials[i].fDiffuse, fMaterials[i].fDiffuse ), 0 );
			pMtl->SetSelfIllum( fMaterials[i].fEmittance,0);
			pMtl->SetOpacity( ( (fMaterials[i].fFlags & tMaterial::kAdditive) ? 0.5f : 1.0f ), 0 );	// $$$ wrong!
			if ( fMaterials[i].fFlags & tMaterial::kTexMapped )
			{
				// Create the diffuse map
				pMtl->SetName( fStringPool.Addr(fMaterials[i].fDiffuseTexture.fTexNameOffset) );
				_makepath( string, fDrive, fDirectory, fStringPool.Addr(fMaterials[i].fDiffuseTexture.fTexNameOffset), ".bmp" );
				if ( _access( string, 0 ) )
				{
					_makepath( string, fDrive, fDirectory, fStringPool.Addr(fMaterials[i].fDiffuseTexture.fTexNameOffset), ".pcx" );
				}
				pMtl->EnableMap( ID_DI, TRUE );
				BitmapTex *bmt = NewDefaultBitmapTex();
				bmt->SetMapName(string);
				pMtl->SetSubTexmap( ID_DI, bmt );
				pMtl->SetActiveTexmap(bmt);
				if ( bmt->SupportTexDisplay() )
				{
					bmt->SetMtlFlag(MTL_TEX_DISPLAY_ENABLED, TRUE);
					bmt->ActivateTexDisplay(TRUE);
					pMtl->SetMtlFlag(MTL_TEX_DISPLAY_ENABLED, TRUE);
				}
				// Create the self-illumination map
				if ( fMaterials[i].fFlags & tMaterial::kLuminance )
				{
					_makepath( string, fDrive, fDirectory, fStringPool.Addr(fMaterials[i].fLuminanceTexture.fTexNameOffset), ".bmp" );
					if ( _access( string, 0 ) )
					{
						_makepath( string, fDrive, fDirectory, fStringPool.Addr(fMaterials[i].fLuminanceTexture.fTexNameOffset), ".pcx" );
					}
					pMtl->EnableMap( ID_SI, TRUE );
					BitmapTex *bmt = NewDefaultBitmapTex();
					bmt->SetMapName(string);
					pMtl->SetSubTexmap( ID_SI, bmt );
					if ( bmt->SupportTexDisplay() )
					{
						bmt->SetMtlFlag(MTL_TEX_DISPLAY_ENABLED, TRUE);
						bmt->ActivateTexDisplay(TRUE);
						pMtl->SetMtlFlag(MTL_TEX_DISPLAY_ENABLED, TRUE);
					}
				}
			}
			else
			{
				pMtl->SetName( "Generic" );
			}

			pMultiMtl->SetSubMtl( i, pMtl );
		}
	}

	// Build the dummy objects
	for ( i = 0; i < fNumNamedPoints; i++ )
	{
		DummyObject* pDummyObject = new DummyObject;
		const float kRad = fFlatHeader.fRadius * 0.025f;
		Box3 box( Point3( -kRad, -kRad, -kRad ), Point3( kRad, kRad, kRad ) );
		pDummyObject->SetBox( box );

		// Create a node to reference the triangle mesh
		ImpNode *pNode = ii->CreateNode();
		if(pNode) 
		{
			Matrix3 tm;
			tm.IdentityMatrix();
			Point3 point( fNamedPoints[i].fPos.x, fNamedPoints[i].fPos.y, fNamedPoints[i].fPos.z );
			tm.Translate( point );
			tm.RotateY(-kPI*0.5f);
			tm.RotateX(kPI*0.5f);
			pNode->Reference(pDummyObject);
			pNode->SetTransform(0,tm);
			pNode->SetName( fStringPool.Addr(fNamedPoints[i].fNameOffset) );
			ii->AddNodeToScene(pNode);
		}
		else
		{
			delete pDummyObject;
		}
	}
	

	// Build triangle meshes
	tMesh* pMesh = fFirstMesh;

	int meshCount = 0;
	while ( pMesh )
	{
		TriObject *pTriObject = CreateNewTriObject();
		if( pTriObject)
		{
			// Load the triangle mesh data

			Matrix3 transMatrix;
			transMatrix.IdentityMatrix();
			transMatrix.RotateY(-kPI*0.5f);
			transMatrix.RotateX(kPI*0.5f);

			Mesh& mesh = pTriObject->mesh;
			mesh.setNumVerts( pMesh->fNumVertices );
			for ( i = 0; i < pMesh->fNumVertices; i++ )
			{
				mesh.verts[i].x = pMesh->fVertices[i].x;
				mesh.verts[i].y = pMesh->fVertices[i].y;
				mesh.verts[i].z = pMesh->fVertices[i].z;
				mesh.verts[i] = mesh.verts[i] * transMatrix;
			}

			// Count the number of triangles;
			int numTriangles = 0;
			for ( i = 0; i < pMesh->fNumFaces; i++ )
			{
				if ( pMesh->fFaces[i].vtxCount > 2 )
				{
					numTriangles += pMesh->fFaces[i].vtxCount - 2;
				}
			}
			// Convert polygons to triangles
			mesh.setNumFaces( numTriangles );
			mesh.setNumTVFaces( numTriangles );
			mesh.setNumTVerts( numTriangles * 3 );	// $$$ inefficient
			int triIndex = 0;
			for ( i = 0; i < pMesh->fNumFaces; i++ )
			{
				if ( pMesh->fFaces[i].vtxCount < 3 )
				{
					continue;
				}
				Point3 faceNormal( pMesh->fFaces[i].normal.x, pMesh->fFaces[i].normal.y, pMesh->fFaces[i].normal.z );
				for ( int j = 1; j < pMesh->fFaces[i].vtxCount - 1; j++ )
				{
					int texIndex = triIndex * 3;
					mesh.faces[triIndex].flags = EDGE_ALL;	
					mesh.faces[triIndex].setMatID((MtlID) pMesh->fFaces[i].materialNdx);
					if ( fMaterials[pMesh->fFaces[i].materialNdx].fFlags & tMaterial::kGouraud )
					{
						mesh.faces[triIndex].smGroup = 1 << pMesh->fFaces[i].materialNdx;	// Convert material to smoothing group
					}
					else
					{
						mesh.faces[triIndex].smGroup = 0;
					}
					mesh.setTVert(texIndex + vIdx[0], pMesh->fFaces[i].vertices[j].u, pMesh->fFaces[i].vertices[j].v, 0 );	// $$$ inefficient
					mesh.setTVert(texIndex + vIdx[1], pMesh->fFaces[i].vertices[j+1].u, pMesh->fFaces[i].vertices[j+1].v, 0 );	// $$$ inefficient
					mesh.setTVert(texIndex + vIdx[2], pMesh->fFaces[i].vertices[0].u, pMesh->fFaces[i].vertices[0].v, 0 );	// $$$ inefficient
					mesh.tvFace[triIndex].t[0] = texIndex;	// $$$ inefficient
					mesh.tvFace[triIndex].t[1] = texIndex + 1;	// $$$ inefficient
					mesh.tvFace[triIndex].t[2] = texIndex + 2;	// $$$ inefficient
					mesh.faces[triIndex].v[vIdx[0]] = pMesh->fFaces[i].vertices[j].vtxNdx;
					mesh.faces[triIndex].v[vIdx[1]] = pMesh->fFaces[i].vertices[j+1].vtxNdx;
					mesh.faces[triIndex++].v[vIdx[2]] = pMesh->fFaces[i].vertices[0].vtxNdx;
				}
			}

			// Create a node to reference the triangle mesh
			ImpNode *pNode = ii->CreateNode();
			if(pNode) 
			{
				Matrix3 tm;
				tm.IdentityMatrix();
				pNode->Reference(pTriObject);
				pNode->SetTransform(0,tm);
				char name[_MAX_FNAME+5];
				sprintf( name, "%s_%d", fFilename, ++meshCount );
				pNode->SetName( name );
				INode* pINode = pNode->GetINode();
				pINode->SetMtl(pMultiMtl);
				ii->AddNodeToScene(pNode);
			}
			else
			{
				delete pTriObject;
			}
		}

		pMesh = pMesh->fNext;
	}
}

void tSFCMODImport::mSplitBreakUpMeshes()
{
	tMesh* newList = NULL, *curMesh = NULL;

	// Counting and new list creation loop
	tMesh* mesh = fFirstMesh;
	if ( mesh )
	{
		newList = curMesh = new tMesh;
	}
	while ( mesh )
	{
		unsigned int v_start = 0, v_stop = 0, f_start = 0;
		for ( unsigned int i = 0; i < mesh->fNumFaces; i++ )
		{
			unsigned int j;
			for ( j = 0; j < mesh->fFaces[i].vtxCount; j++ )
			{
				if ( v_start > mesh->fFaces[i].vertices[j].vtxNdx )
				{
					v_start = mesh->fFaces[i].vertices[j].vtxNdx;
				}
				if ( v_stop < mesh->fFaces[i].vertices[j].vtxNdx )
				{
					v_stop = mesh->fFaces[i].vertices[j].vtxNdx;
				}
			}

			// Add new group
			if ( i + 1 >= mesh->fNumFaces || mesh->fFaces[i+1].vertices[0].flags != mesh->fFaces[i].vertices[0].flags )
			{
				// Load the group with the mesh data
				curMesh->fNumFaces = i - f_start + 1;
				curMesh->fFaces = new tFaceDef[curMesh->fNumFaces];
				curMesh->fNumVertices = v_stop - v_start + 1;
				curMesh->fVertices = new tPoint3[curMesh->fNumVertices];
				for ( j = 0; j < curMesh->fNumVertices; j++ )
				{
					curMesh->fVertices[j] = mesh->fVertices[ v_start + j ];
				}
				for ( j = 0; j < curMesh->fNumFaces; j++ )
				{
					curMesh->fFaces[j] = mesh->fFaces[ f_start + j ];
					curMesh->fFaces[j].vertices = new tFaceDef::tFlatPolyVertex[ curMesh->fFaces[j].vtxCount ];
					for ( unsigned int k = 0; k < curMesh->fFaces[j].vtxCount; k++ )
					{
						curMesh->fFaces[j].vertices[k] = mesh->fFaces[ f_start + j ].vertices[k];
						curMesh->fFaces[j].vertices[k].vtxNdx -= v_start;
					}
				}

				v_start = v_stop + 1;
				v_stop = v_start;
				f_start = i + 1;
				if ( i + 1 < mesh->fNumFaces || mesh->fNext )
				{
					curMesh->fNext = new tMesh;
					curMesh = curMesh->fNext;
				}
			}
		}
		mesh = mesh->fNext;
	}

	// Remove old mesh contents
	curMesh = fFirstMesh;
	while ( curMesh )
	{
		fFirstMesh = curMesh->fNext;
		delete curMesh;
		curMesh = fFirstMesh;
	}

	fFirstMesh = newList;
}

int tSFCMODImport::mLoadSFCMOD( const TCHAR* filename, ImpInterface* ii, Interface* gi )
{
	bool bReplaceScene = ( MessageBox( IDS_SD_REPLACESCENE, IDS_SFCMODEXP, MB_YESNO ) == IDYES );

	_splitpath( filename, fDrive, fDirectory, fFilename, NULL );

	// Open a file stream
	FILE* myFile = fopen( filename, "rb" );

	fpos_t endOfFile;
	fseek( myFile, 0, SEEK_END );
	fgetpos( myFile, &endOfFile );
	fseek( myFile, 0, SEEK_SET );

	mReadChunk( myFile, endOfFile );

	fclose( myFile );

	if ( MessageBox( IDS_SD_BREAKUPPROMPT, IDS_SFCMODIMP, MB_YESNO ) == IDYES )
	{
		mSplitBreakUpMeshes();
	}
	
	if ( bReplaceScene )
	{
		ii->NewScene();
	}

	mBuildScene( ii, gi );

	ii->RedrawViews();

	return 1;
}

int tSFCMODImport::DoImport(const TCHAR *filename,ImpInterface *ii,Interface *gi, BOOL suppressPrompts) 
{
//	DialogBox( hInstance, MAKEINTRESOURCE( IDD_GREETING ), GetActiveWindow(), MsgDlgProc );
	// Set a global prompt display switch
	showPrompts = suppressPrompts ? FALSE : TRUE;

	int status;

	status = mLoadSFCMOD( filename, ii, gi );

	return IMPEXP_SUCCESS;
}


